@@ -1,33 +1,10 @@
 #region Copyright notice and license
 // Protocol Buffers - Google's data interchange format
 // Copyright 2008 Google Inc.  All rights reserved.
-// https://developers.google.com/protocol-buffers/
 //
-// Redistribution and use in source and binary forms, with or without
-// modification, are permitted provided that the following conditions are
-// met:
-//
-//     * Redistributions of source code must retain the above copyright
-// notice, this list of conditions and the following disclaimer.
-//     * Redistributions in binary form must reproduce the above
-// copyright notice, this list of conditions and the following disclaimer
-// in the documentation and/or other materials provided with the
-// distribution.
-//     * Neither the name of Google Inc. nor the names of its
-// contributors may be used to endorse or promote products derived from
-// this software without specific prior written permission.
-//
-// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
-// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
-// LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
-// A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
-// OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
-// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
-// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
-// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
-// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
-// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
-// OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
+// Use of this source code is governed by a BSD-style
+// license that can be found in the LICENSE file or at
+// https://developers.google.com/open-source/licenses/bsd
 #endregion
 
 using Google.Protobuf.Collections;
@@ -44,8 +21,6 @@ namespace Google.Protobuf.Reflection
         private EnumDescriptor enumType;
         private MessageDescriptor extendeeType;
         private MessageDescriptor messageType;
-        private FieldType fieldType;
-        private readonly string propertyName; // Annoyingly, needed in Crosslink.
         private IFieldAccessor accessor;
 
         /// <summary>
@@ -70,6 +45,11 @@ namespace Google.Protobuf.Reflection
         /// </summary>
         public string JsonName { get; }
 
+        /// <summary>
+        /// The name of the property in the <c>ContainingType.ClrType</c> class.
+        /// </summary>
+        public string PropertyName { get; }
+
         /// <summary>
         /// Indicates whether this field supports presence, either implicitly (e.g. due to it being a message
         /// type field) or explicitly via Has/Clear members. If this returns true, it is safe to call
@@ -81,12 +61,21 @@ namespace Google.Protobuf.Reflection
             : IsRepeated ? false
             : IsMap ? false
             : FieldType == FieldType.Message ? true
+            : FieldType == FieldType.Group ? true
             // This covers "real oneof members" and "proto3 optional fields"
             : ContainingOneof != null ? true
-            : File.Syntax == Syntax.Proto2;
+            : Features.FieldPresence != FeatureSet.Types.FieldPresence.Implicit;
 
         internal FieldDescriptorProto Proto { get; }
 
+        /// <summary>
+        /// Returns a clone of the underlying <see cref="FieldDescriptorProto"/> describing this field.
+        /// Note that a copy is taken every time this method is called, so clients using it frequently
+        /// (and not modifying it) may want to cache the returned value.
+        /// </summary>
+        /// <returns>A protobuf representation of this field descriptor.</returns>
+        public FieldDescriptorProto ToProto() => Proto.Clone();
+
         /// <summary>
         /// An extension identifier for this field, or <c>null</c> if this field isn't an extension.
         /// </summary>
@@ -94,12 +83,18 @@ namespace Google.Protobuf.Reflection
 
         internal FieldDescriptor(FieldDescriptorProto proto, FileDescriptor file,
                                  MessageDescriptor parent, int index, string propertyName, Extension extension)
-            : base(file, file.ComputeFullName(parent, proto.Name), index)
+            : base(file, file.ComputeFullName(parent, proto.Name), index,
+                  GetDirectParentFeatures(proto, file, parent).MergedWith(InferFeatures(file, proto)).MergedWith(proto.Options?.Features))
         {
             Proto = proto;
-            if (proto.Type != 0)
+            if (proto.HasType)
             {
-                fieldType = GetFieldTypeFromProtoType(proto.Type);
+                FieldType = GetFieldTypeFromProtoType(proto.Type);
+                if (FieldType == FieldType.Message &&
+                    Features.MessageEncoding == FeatureSet.Types.MessageEncoding.Delimited)
+                {
+                    FieldType = FieldType.Group;
+                }
             }
 
             if (FieldNumber <= 0)
@@ -123,11 +118,57 @@ namespace Google.Protobuf.Reflection
             // for later.
             // We could trust the generated code and check whether the type of the property is
             // a MapField, but that feels a tad nasty.
-            this.propertyName = propertyName;
+            PropertyName = propertyName;
             Extension = extension;
             JsonName =  Proto.JsonName == "" ? JsonFormatter.ToJsonName(Proto.Name) : Proto.JsonName;
         }
 
+        /// <summary>
+        /// Returns the features from the direct parent:
+        /// - The file for top-level extensions
+        /// - The oneof for one-of fields
+        /// - Otherwise the message
+        /// </summary>
+        private static FeatureSetDescriptor GetDirectParentFeatures(FieldDescriptorProto proto, FileDescriptor file, MessageDescriptor parent) =>
+            parent is null ? file.Features
+            // Ignore invalid oneof indexes here; they'll be validated later anyway.
+            : proto.OneofIndex >= 0 && proto.OneofIndex < parent.Proto.OneofDecl.Count ? parent.Oneofs[proto.OneofIndex].Features
+            : parent.Features;
+
+        /// <summary>
+        /// Returns a feature set with inferred features for the given field, or null if no features
+        /// need to be inferred.
+        /// </summary>
+        private static FeatureSet InferFeatures(FileDescriptor file, FieldDescriptorProto proto)
+        {
+            if ((int) file.Edition >= (int) Edition._2023)
+            {
+                return null;
+            }
+            // This is lazily initialized, as most fields won't need it.
+            FeatureSet features = null;
+            if (proto.Label == FieldDescriptorProto.Types.Label.Required)
+            {
+                features ??= new FeatureSet();
+                features.FieldPresence = FeatureSet.Types.FieldPresence.LegacyRequired;
+            }
+            if (proto.Type == FieldDescriptorProto.Types.Type.Group)
+            {
+                features ??= new FeatureSet();
+                features.MessageEncoding = FeatureSet.Types.MessageEncoding.Delimited;
+            }
+            if (file.Edition == Edition.Proto2 && (proto.Options?.Packed ?? false))
+            {
+                features ??= new FeatureSet();
+                features.RepeatedFieldEncoding = FeatureSet.Types.RepeatedFieldEncoding.Packed;
+            }
+            if (file.Edition == Edition.Proto3 && !(proto.Options?.Packed ?? true))
+            {
+                features ??= new FeatureSet();
+                features.RepeatedFieldEncoding = FeatureSet.Types.RepeatedFieldEncoding.Expanded;
+            }
+            return features;
+        }
 
         /// <summary>
         /// The brief name of the descriptor's target.
@@ -164,47 +205,28 @@ namespace Google.Protobuf.Reflection
         /// </summary>
         private static FieldType GetFieldTypeFromProtoType(FieldDescriptorProto.Types.Type type)
         {
-            switch (type)
+            return type switch
             {
-                case FieldDescriptorProto.Types.Type.Double:
-                    return FieldType.Double;
-                case FieldDescriptorProto.Types.Type.Float:
-                    return FieldType.Float;
-                case FieldDescriptorProto.Types.Type.Int64:
-                    return FieldType.Int64;
-                case FieldDescriptorProto.Types.Type.Uint64:
-                    return FieldType.UInt64;
-                case FieldDescriptorProto.Types.Type.Int32:
-                    return FieldType.Int32;
-                case FieldDescriptorProto.Types.Type.Fixed64:
-                    return FieldType.Fixed64;
-                case FieldDescriptorProto.Types.Type.Fixed32:
-                    return FieldType.Fixed32;
-                case FieldDescriptorProto.Types.Type.Bool:
-                    return FieldType.Bool;
-                case FieldDescriptorProto.Types.Type.String:
-                    return FieldType.String;
-                case FieldDescriptorProto.Types.Type.Group:
-                    return FieldType.Group;
-                case FieldDescriptorProto.Types.Type.Message:
-                    return FieldType.Message;
-                case FieldDescriptorProto.Types.Type.Bytes:
-                    return FieldType.Bytes;
-                case FieldDescriptorProto.Types.Type.Uint32:
-                    return FieldType.UInt32;
-                case FieldDescriptorProto.Types.Type.Enum:
-                    return FieldType.Enum;
-                case FieldDescriptorProto.Types.Type.Sfixed32:
-                    return FieldType.SFixed32;
-                case FieldDescriptorProto.Types.Type.Sfixed64:
-                    return FieldType.SFixed64;
-                case FieldDescriptorProto.Types.Type.Sint32:
-                    return FieldType.SInt32;
-                case FieldDescriptorProto.Types.Type.Sint64:
-                    return FieldType.SInt64;
-                default:
-                    throw new ArgumentException("Invalid type specified");
-            }
+                FieldDescriptorProto.Types.Type.Double => FieldType.Double,
+                FieldDescriptorProto.Types.Type.Float => FieldType.Float,
+                FieldDescriptorProto.Types.Type.Int64 => FieldType.Int64,
+                FieldDescriptorProto.Types.Type.Uint64 => FieldType.UInt64,
+                FieldDescriptorProto.Types.Type.Int32 => FieldType.Int32,
+                FieldDescriptorProto.Types.Type.Fixed64 => FieldType.Fixed64,
+                FieldDescriptorProto.Types.Type.Fixed32 => FieldType.Fixed32,
+                FieldDescriptorProto.Types.Type.Bool => FieldType.Bool,
+                FieldDescriptorProto.Types.Type.String => FieldType.String,
+                FieldDescriptorProto.Types.Type.Group => FieldType.Group,
+                FieldDescriptorProto.Types.Type.Message => FieldType.Message,
+                FieldDescriptorProto.Types.Type.Bytes => FieldType.Bytes,
+                FieldDescriptorProto.Types.Type.Uint32 => FieldType.UInt32,
+                FieldDescriptorProto.Types.Type.Enum => FieldType.Enum,
+                FieldDescriptorProto.Types.Type.Sfixed32 => FieldType.SFixed32,
+                FieldDescriptorProto.Types.Type.Sfixed64 => FieldType.SFixed64,
+                FieldDescriptorProto.Types.Type.Sint32 => FieldType.SInt32,
+                FieldDescriptorProto.Types.Type.Sint64 => FieldType.SInt64,
+                _ => throw new ArgumentException("Invalid type specified"),
+            };
         }
 
         /// <summary>
@@ -215,30 +237,17 @@ namespace Google.Protobuf.Reflection
         /// <summary>
         /// Returns <c>true</c> if this field is a required field; <c>false</c> otherwise.
         /// </summary>
-        public bool IsRequired => Proto.Label == FieldDescriptorProto.Types.Label.Required;
+        public bool IsRequired => Features.FieldPresence == FeatureSet.Types.FieldPresence.LegacyRequired;
 
         /// <summary>
         /// Returns <c>true</c> if this field is a map field; <c>false</c> otherwise.
         /// </summary>
-        public bool IsMap => fieldType == FieldType.Message && messageType.Proto.Options != null && messageType.Proto.Options.MapEntry;
+        public bool IsMap => FieldType == FieldType.Message && messageType.Proto.Options != null && messageType.Proto.Options.MapEntry;
 
         /// <summary>
         /// Returns <c>true</c> if this field is a packed, repeated field; <c>false</c> otherwise.
         /// </summary>
-        public bool IsPacked
-        {
-            get
-            {
-                if (File.Syntax != Syntax.Proto3)
-                {
-                    return Proto.Options?.Packed ?? false;
-                }
-                else
-                {
-                    return !Proto.Options.HasPacked || Proto.Options.Packed;
-                }
-            }
-        }
+        public bool IsPacked => Features.RepeatedFieldEncoding == FeatureSet.Types.RepeatedFieldEncoding.Packed;
 
         /// <summary>
         /// Returns <c>true</c> if this field extends another message type; <c>false</c> otherwise.
@@ -248,7 +257,7 @@ namespace Google.Protobuf.Reflection
         /// <summary>
         /// Returns the type of the field.
         /// </summary>
-        public FieldType FieldType => fieldType;
+        public FieldType FieldType { get; private set; }
 
         /// <summary>
         /// Returns the field number declared in the proto file.
@@ -278,7 +287,7 @@ namespace Google.Protobuf.Reflection
         {
             get
             {
-                if (fieldType != FieldType.Enum)
+                if (FieldType != FieldType.Enum)
                 {
                     throw new InvalidOperationException("EnumType is only valid for enum fields.");
                 }
@@ -293,7 +302,7 @@ namespace Google.Protobuf.Reflection
         {
             get
             {
-                if (fieldType != FieldType.Message && fieldType != FieldType.Group)
+                if (FieldType != FieldType.Message && FieldType != FieldType.Group)
                 {
                     throw new InvalidOperationException("MessageType is only valid for message or group fields.");
                 }
@@ -328,12 +337,23 @@ namespace Google.Protobuf.Reflection
         /// Custom options can be retrieved as extensions of the returned message.
         /// NOTE: A defensive copy is created each time this property is retrieved.
         /// </summary>
-        public FieldOptions GetOptions() => Proto.Options?.Clone();
+        public FieldOptions GetOptions()
+        {
+            var clone = Proto.Options?.Clone();
+            if (clone is null)
+            {
+                return null;
+            }
+            // Clients should be using feature accessor methods, not accessing features on the
+            // options proto.
+            clone.Features = null;
+            return clone;
+        }
 
         /// <summary>
         /// Gets a single value field option for this descriptor
         /// </summary>
-         [Obsolete("GetOption is obsolete. Use the GetOptions() method.")]
+        [Obsolete("GetOption is obsolete. Use the GetOptions() method.")]
         public T GetOption<T>(Extension<FieldOptions, T> extension)
         {
             var value = Proto.Options.GetExtension(extension);
@@ -359,16 +379,20 @@ namespace Google.Protobuf.Reflection
                 IDescriptor typeDescriptor =
                     File.DescriptorPool.LookupSymbol(Proto.TypeName, this);
 
-                if (Proto.HasType)
+                // TODO: See how much of this is actually required.
+                if (!Proto.HasType)
                 {
                     // Choose field type based on symbol.
                     if (typeDescriptor is MessageDescriptor)
                     {
-                        fieldType = FieldType.Message;
+                        FieldType =
+                            Features.MessageEncoding == FeatureSet.Types.MessageEncoding.Delimited
+                            ? FieldType.Group
+                            : FieldType.Message;
                     }
                     else if (typeDescriptor is EnumDescriptor)
                     {
-                        fieldType = FieldType.Enum;
+                        FieldType = FieldType.Enum;
                     }
                     else
                     {
@@ -376,26 +400,26 @@ namespace Google.Protobuf.Reflection
                     }
                 }
 
-                if (fieldType == FieldType.Message || fieldType == FieldType.Group)
+                if (FieldType == FieldType.Message || FieldType == FieldType.Group)
                 {
-                    if (!(typeDescriptor is MessageDescriptor))
+                    if (typeDescriptor is not MessageDescriptor m)
                     {
                         throw new DescriptorValidationException(this, $"\"{Proto.TypeName}\" is not a message type.");
                     }
-                    messageType = (MessageDescriptor) typeDescriptor;
+                    messageType = m;
 
                     if (Proto.HasDefaultValue)
                     {
                         throw new DescriptorValidationException(this, "Messages can't have default values.");
                     }
                 }
-                else if (fieldType == FieldType.Enum)
+                else if (FieldType == FieldType.Enum)
                 {
-                    if (!(typeDescriptor is EnumDescriptor))
+                    if (typeDescriptor is not EnumDescriptor e)
                     {
                         throw new DescriptorValidationException(this, $"\"{Proto.TypeName}\" is not an enum type.");
                     }
-                    enumType = (EnumDescriptor) typeDescriptor;
+                    enumType = e;
                 }
                 else
                 {
@@ -404,7 +428,7 @@ namespace Google.Protobuf.Reflection
             }
             else
             {
-                if (fieldType == FieldType.Message || fieldType == FieldType.Enum)
+                if (FieldType == FieldType.Message || FieldType == FieldType.Enum)
                 {
                     throw new DescriptorValidationException(this, "Field with message or enum type missing type_name.");
                 }
@@ -436,20 +460,19 @@ namespace Google.Protobuf.Reflection
             // If we're given no property name, that's because we really don't want an accessor.
             // This could be because it's a map message, or it could be that we're loading a FileDescriptor dynamically.
             // TODO: Support dynamic messages.
-            if (propertyName == null)
+            if (PropertyName == null)
             {
                 return null;
             }
 
-            var property = ContainingType.ClrType.GetProperty(propertyName);
+            var property = ContainingType.ClrType.GetProperty(PropertyName);
             if (property == null)
             {
-                throw new DescriptorValidationException(this, $"Property {propertyName} not found in {ContainingType.ClrType}");
+                throw new DescriptorValidationException(this, $"Property {PropertyName} not found in {ContainingType.ClrType}");
             }
             return IsMap ? new MapFieldAccessor(property, this)
                 : IsRepeated ? new RepeatedFieldAccessor(property, this)
-                : (IFieldAccessor) new SingleFieldAccessor(property, this);
+                : (IFieldAccessor) new SingleFieldAccessor(ContainingType.ClrType, property, this);
         }
     }
 }
- 
\ No newline at end of file
