@@ -1,37 +1,18 @@
 // Protocol Buffers - Google's data interchange format
 // Copyright 2008 Google Inc.  All rights reserved.
-// https://developers.google.com/protocol-buffers/
 //
-// Redistribution and use in source and binary forms, with or without
-// modification, are permitted provided that the following conditions are
-// met:
-//
-//     * Redistributions of source code must retain the above copyright
-// notice, this list of conditions and the following disclaimer.
-//     * Redistributions in binary form must reproduce the above
-// copyright notice, this list of conditions and the following disclaimer
-// in the documentation and/or other materials provided with the
-// distribution.
-//     * Neither the name of Google Inc. nor the names of its
-// contributors may be used to endorse or promote products derived from
-// this software without specific prior written permission.
-//
-// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
-// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
-// LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
-// A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
-// OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
-// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
-// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
-// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
-// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
-// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
-// OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
+// Use of this source code is governed by a BSD-style
+// license that can be found in the LICENSE file or at
+// https://developers.google.com/open-source/licenses/bsd
+
+#include <cstdint>
+#include <string>
 
-#include <gmock/gmock.h>
 #include <gtest/gtest.h>
 #include "absl/log/absl_check.h"
 #include "google/protobuf/descriptor.h"
+#include "google/protobuf/generated_message_bases.h"
+#include "google/protobuf/repeated_ptr_field.h"
 #include "google/protobuf/unittest.pb.h"
 
 // Must be included last.
@@ -121,6 +102,8 @@ TEST(GeneratedMessageTest, EmptyMessageWithExtensionsSize) {
 }
 
 TEST(GeneratedMessageTest, RecursiveMessageSize) {
+  // TODO: remove once synthetic_pdproto lands.
+#ifndef PROTOBUF_FORCE_SPLIT
   struct MockGenerated : public MockMessageBase {  // 16 bytes
     int has_bits[1];                               // 4 bytes
     int cached_size;                               // 4 bytes
@@ -130,6 +113,17 @@ TEST(GeneratedMessageTest, RecursiveMessageSize) {
     // + 0-4 bytes padding
   };
   ABSL_CHECK_MESSAGE_SIZE(MockGenerated, 40);
+#else   // !PROTOBUF_FORCE_SPLIT
+  struct MockGenerated : public MockMessageBase {  // 16 bytes
+    int has_bits[1];                               // 4 bytes
+    int cached_size;                               // 4 bytes
+    void* split;                                   // 8 bytes
+    PROTOBUF_TSAN_DECLARE_MEMBER;                  // 0-4 bytes
+    // + 0-4 bytes padding
+  };
+  ABSL_CHECK_MESSAGE_SIZE(MockGenerated, 32);
+#endif  // PROTOBUF_FORCE_SPLIT
+
   EXPECT_EQ(sizeof(protobuf_unittest::TestRecursiveMessage),
             sizeof(MockGenerated));
 }
@@ -139,7 +133,7 @@ TEST(GeneratedMessageTest, OneStringSize) {
     int has_bits[1];                               // 4 bytes
     int cached_size;                               // 4 bytes
     PROTOBUF_TSAN_DECLARE_MEMBER;                  // 0-4 bytes
-    // + 0-4 bytes padding
+                                                   // + 0-4 bytes padding
     void* data;                                    // 8 bytes
   };
   ABSL_CHECK_MESSAGE_SIZE(MockGenerated, 32);
@@ -147,13 +141,24 @@ TEST(GeneratedMessageTest, OneStringSize) {
 }
 
 TEST(GeneratedMessageTest, MoreStringSize) {
+  // TODO: remove once synthetic_pdproto lands.
+#ifndef PROTOBUF_FORCE_SPLIT
   struct MockGenerated : public MockMessageBase {  // 16 bytes
     int cached_size;                               // 4 bytes
     PROTOBUF_TSAN_DECLARE_MEMBER;                  // 0-4 bytes
-    // + 0-4 bytes padding
+                                                   // + 0-4 bytes padding
     MockRepeatedPtrField data;                     // 24 bytes
   };
   ABSL_CHECK_MESSAGE_SIZE(MockGenerated, 48);
+#else   // !PROTOBUF_FORCE_SPLIT
+  struct MockGenerated : public MockMessageBase {  // 16 bytes
+    int cached_size;                               // 4 bytes
+    void* split;                                   // 8 bytes
+    PROTOBUF_TSAN_DECLARE_MEMBER;                  // 0-4 bytes
+                                                   // + 0-4 bytes padding
+  };
+  ABSL_CHECK_MESSAGE_SIZE(MockGenerated, 32);
+#endif  // PROTOBUF_FORCE_SPLIT
   EXPECT_EQ(sizeof(protobuf_unittest::MoreString), sizeof(MockGenerated));
 }
 
@@ -162,7 +167,7 @@ TEST(GeneratedMessageTest, Int32MessageSize) {
     int has_bits[1];                               // 4 bytes
     int cached_size;                               // 4 bytes
     PROTOBUF_TSAN_DECLARE_MEMBER;                  // 0-4 bytes
-    // + 0-4 bytes padding
+                                                   // + 0-4 bytes padding
     int32_t data;                                  // 4 bytes
   };
   ABSL_CHECK_MESSAGE_SIZE(MockGenerated, 32);
@@ -174,7 +179,7 @@ TEST(GeneratedMessageTest, Int64MessageSize) {
     int has_bits[1];                               // 4 bytes
     int cached_size;                               // 4 bytes
     PROTOBUF_TSAN_DECLARE_MEMBER;                  // 0-4 bytes
-    // + 0-4 bytes padding
+                                                   // + 0-4 bytes padding
     int64_t data;                                  // 8 bytes
   };
   ABSL_CHECK_MESSAGE_SIZE(MockGenerated, 32);
@@ -186,7 +191,7 @@ TEST(GeneratedMessageTest, BoolMessageSize) {
     int has_bits[1];                               // 4 bytes
     int cached_size;                               // 4 bytes
     PROTOBUF_TSAN_DECLARE_MEMBER;                  // 0-4 bytes
-    // + 0-4 bytes padding
+                                                   // + 0-4 bytes padding
     bool data;                                     // 1 byte
     // + 3 bytes padding
   };
@@ -199,7 +204,7 @@ TEST(GeneratedMessageTest, OneofSize) {
     void* foo;                                     // 8 bytes
     int cached_size;                               // 4 bytes
     PROTOBUF_TSAN_DECLARE_MEMBER;                  // 0-4 bytes
-    // + 0-4 bytes padding
+                                                   // + 0-4 bytes padding
     uint32_t oneof_case[1];                        // 4 bytes
   };
   ABSL_CHECK_MESSAGE_SIZE(MockGenerated, 32);
@@ -207,11 +212,13 @@ TEST(GeneratedMessageTest, OneofSize) {
 }
 
 TEST(GeneratedMessageTest, Oneof2Size) {
+  // TODO: remove once synthetic_pdproto lands.
+#ifndef PROTOBUF_FORCE_SPLIT
   struct MockGenerated : public MockMessageBase {  // 16 bytes
     int has_bits[1];                               // 4 bytes
     int cached_size;                               // 4 bytes
     PROTOBUF_TSAN_DECLARE_MEMBER;                  // 0-4 bytes
-    // + 0-4 bytes padding
+                                                   // + 0-4 bytes padding
     void* baz_string;                              // 8 bytes
     int32_t baz_int;                               // 4 bytes
                                                    // + 4 bytes padding
@@ -220,10 +227,25 @@ TEST(GeneratedMessageTest, Oneof2Size) {
     uint32_t oneof_case[2];                        // 8 bytes
   };
   ABSL_CHECK_MESSAGE_SIZE(MockGenerated, 64);
+#else   // !PROTOBUF_FORCE_SPLIT
+  struct MockGenerated : public MockMessageBase {  // 16 bytes
+    int has_bits[1];                               // 4 bytes
+    int cached_size;                               // 4 bytes
+    PROTOBUF_TSAN_DECLARE_MEMBER;                  // 0-4 bytes
+                                                   // + 0-4 bytes padding
+    void* split;                                   // 8 bytes
+    void* foo;                                     // 8 bytes
+    void* bar;                                     // 8 bytes
+    uint32_t oneof_case[2];                        // 8 bytes
+  };
+  ABSL_CHECK_MESSAGE_SIZE(MockGenerated, 56);
+#endif  // PROTOBUF_FORCE_SPLIT
   EXPECT_EQ(sizeof(protobuf_unittest::TestOneof2), sizeof(MockGenerated));
 }
 
 TEST(GeneratedMessageTest, FieldOrderingsSize) {
+  // TODO: remove once synthetic_pdproto lands.
+#ifndef PROTOBUF_FORCE_SPLIT
   struct MockGenerated : public MockMessageBase {  // 16 bytes
     int has_bits[1];                               // 4 bytes
     int cached_size;                               // 4 bytes
@@ -236,16 +258,29 @@ TEST(GeneratedMessageTest, FieldOrderingsSize) {
     // + 0-4 bytes padding
   };
   ABSL_CHECK_MESSAGE_SIZE(MockGenerated, 80);
+#else   // !PROTOBUF_FORCE_SPLIT
+  struct MockGenerated : public MockMessageBase {  // 16 bytes
+    int has_bits[1];                               // 4 bytes
+    int cached_size;                               // 4 bytes
+    MockExtensionSet extensions;                   // 24 bytes
+    void* split;                                   // 8 bytes
+    PROTOBUF_TSAN_DECLARE_MEMBER;                  // 0-4 bytes
+    // + 0-4 bytes padding
+  };
+  ABSL_CHECK_MESSAGE_SIZE(MockGenerated, 56);
+#endif  // PROTOBUF_FORCE_SPLIT
   EXPECT_EQ(sizeof(protobuf_unittest::TestFieldOrderings), sizeof(MockGenerated));
 }
 
 TEST(GeneratedMessageTest, TestMessageSize) {
   // We expect the message to contain (not in this order):
+  // TODO: remove once synthetic_pdproto lands.
+#ifndef PROTOBUF_FORCE_SPLIT
   struct MockGenerated : public MockMessageBase {  // 16 bytes
     int has_bits[1];                               // 4 bytes
     int cached_size;                               // 4 bytes
     PROTOBUF_TSAN_DECLARE_MEMBER;                  // 0-4 bytes
-    // + 0-4 bytes padding
+                                                   // + 0-4 bytes padding
     void* m4;                                      // 8 bytes
     int64_t m2;                                    // 8 bytes
     bool m1;                                       // 1 bytes
@@ -255,10 +290,22 @@ TEST(GeneratedMessageTest, TestMessageSize) {
     int64_t m6;                                    // 8 bytes
   };
   ABSL_CHECK_MESSAGE_SIZE(MockGenerated, 56);
+#else   // !PROTOBUF_FORCE_SPLIT
+  struct MockGenerated : public MockMessageBase {  // 16 bytes
+    int has_bits[1];                               // 4 bytes
+    int cached_size;                               // 4 bytes
+    void* split;                                   // 8 bytes
+    PROTOBUF_TSAN_DECLARE_MEMBER;                  // 0-4 bytes
+                                                   // + 0-4 bytes padding
+  };
+  ABSL_CHECK_MESSAGE_SIZE(MockGenerated, 32);
+#endif  // PROTOBUF_FORCE_SPLIT
   EXPECT_EQ(sizeof(protobuf_unittest::TestMessageSize), sizeof(MockGenerated));
 }
 
 TEST(GeneratedMessageTest, PackedTypesSize) {
+  // TODO: remove once synthetic_pdproto lands.
+#ifndef PROTOBUF_FORCE_SPLIT
   struct MockGenerated : public MockMessageBase {  // 16 bytes
     MockRepeatedField packed_int32;                // 16 bytes
     int packed_int32_cached_byte_size;             // 4 bytes + 4 bytes padding
@@ -286,6 +333,15 @@ TEST(GeneratedMessageTest, PackedTypesSize) {
     // + 0-4 bytes padding
   };
   ABSL_CHECK_MESSAGE_SIZE(MockGenerated, 16 * 15 + 8 * 6 + 8);
+#else   // !PROTOBUF_FORCE_SPLIT
+  struct MockGenerated : public MockMessageBase {  // 16 bytes
+    int cached_size;                               // 4 bytes + 4 bytes padding
+    void* split;                                   // 8 bytes
+    PROTOBUF_TSAN_DECLARE_MEMBER;                  // 0-4 bytes
+    // + 0-4 bytes padding
+  };
+  ABSL_CHECK_MESSAGE_SIZE(MockGenerated, 32);
+#endif  // PROTOBUF_FORCE_SPLIT
   EXPECT_EQ(sizeof(protobuf_unittest::TestPackedTypes), sizeof(MockGenerated));
 }
 
